%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Code to Draw Fig. 1 in the main manuscript
%
% O. A. Elgendy and S. H. Chan,
% "Optimal Threshold Design for Quanta Image Sensor,"
% IEEE Transactions on Computational Imaging, 2017.
% http://ieeexplore.ieee.org/document/8169093/
%
% Omar Elgendy and Stanley Chan
% Copyright 2017
% Purdue University, West Lafayette, In, USA.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

clc,clear,close all
rng('default');

K                    =  4;      	     %  Spatial  Overasampling Factor
T_adapt              =  8;      	     %  Frames for Threshold Adaptation
T_recon              =  22;     	     %  Frames for Image Recosntruction
T                    =  T_adapt+T_recon; %  Temporal Overasampling Factor
Qblock               =  1;               %  1x1 pixels share the same threshold
qmax                 =  16;              %  Maximum Quantization Theshold
g                    =  ones(K)/K^2;     %  Interpolation kernel
scale                =  0.2;             %  Downscale the original image

% option variable for the "generateQmap" function
opts.qmax            =  qmax;            
opts.T_adapt         =  T_adapt;

IM                   =  im2double(imread('.\ImageData\SSIPDataset\IMG_0170.JPG'));
if size(IM,3)~=1;IM  =  rgb2gray(IM);end
IM                   =  imresize(IM,scale);
[rows,cols,NbChs]    =  size(IM);

%------------------QIS Simulation------------------------------------------
c                    =  max(0,IM);
alpha                =  K^2*qmax;  % Sensor Gain
oracleQmap           =  min(floor(alpha*c/K^2)+1,qmax);
opts.oracleQmap      =  oracleQmap;
y                    =  generatePhotonCounts(c,K,T,alpha,g);
IMGT                 =  c;
%---------------Uniform Qmap low threshold---------------------------------
ql                   =  3;
opts.q               =  ql;
[QmapLR,QmapHR]      =  generateQmap([rows,cols],K,'uniform',opts);
b                    =  double(y>=repmat(QmapHR,1,1,T));
IMql                 =  imageReconstruct(b,K,alpha,QmapLR);
PSNRq1               =  psnr(IMql,c);
fprintf(' PSNR_q1 %2.0f dB\n',PSNRq1);
%---------------Uniform Qmap high threshold--------------------------------
qh                   =  12;
opts.q               =  qh;
[QmapLR,QmapHR]      =  generateQmap([rows,cols],K,'uniform',opts);
b                    =  double(y>=repmat(QmapHR,1,1,T));
IMqh                 =  imageReconstruct(b,K,alpha,QmapLR);
PSNRqh               =  psnr(IMqh,c);
fprintf(' PSNR_qh %2.0f dB\n',PSNRqh);
%---------------Uniform Qmap medium threshold------------------------------
qm                   =  qmax/2;
opts.q               =  qm;
[QmapLR,QmapHR]      =  generateQmap([rows,cols],K,'uniform',opts);
b                    =  double(y>=repmat(QmapHR,1,1,T));
IMqm                 =  imageReconstruct(b,K,alpha,QmapLR);
PSNRqm               =  psnr(IMqm,c);
fprintf(' PSNR_qm %2.0f dB\n',PSNRqm);
%-------------------KxK Adapted Qmap---------------------------------------
opts.y               =  y;
opts.Qblock          =  Qblock;
[QmapLR,QmapHR]      =  generateQmap([rows,cols],K,'bisection',opts);
framesqs             =  T_adapt + (1:T_recon);
b                    =  double(y(:,:,framesqs)>=repmat(QmapHR,1,1,T_recon));
IMqs                 =  imageReconstruct(b,K,alpha,QmapLR);
PSNRqs               =  psnr(IMqs,IM);
fprintf(' PSNR_qs %2.0f dB /',PSNRqs);
%-------------Simulate 1 frame of photon counts----------------------------
c                    =  max(0,IM);      %figure;hist(10*log10(c0(:)),100)
alpha                =  K^2*qmax/max(c(:));
theta                =  alpha*kron(c,ones(K))/K^2;
y1                   =  poissrnd(theta);

%% Draw Results
TLx1  =  418;  
TLy1  =  495;
Kx    =  50; 
Ky    =  50;
pos1  =  [0.13 0.625 0.3 0.3];
scale =  1;

b  =  1*(y1>=ql);
figure;imagesc(b(K*TLy1:K*(TLy1+Ky),K*TLx1:K*(TLx1+Kx)));
set(gca,'xtick',[],'ytick',[],'LineWidth',2,'XColor',[1 0 0],'YColor',[1 0 0]);
colormap(gray(256));box on;%axis off;

b  =  1*(y1>=qh);
figure;imagesc(imresize(b(K*TLy1:K*(TLy1+Ky),K*TLx1:K*(TLx1+Kx)),scale,'nearest'));
set(gca,'xtick',[],'ytick',[],'LineWidth',2,'XColor',[1 0 0],'YColor',[1 0 0]);
colormap(gray(256));box on;%axis off;

b  =  1*(y1>=QmapHR);
figure;imagesc(imresize(b(K*TLy1:K*(TLy1+Ky),K*TLx1:K*(TLx1+Kx)),scale,'nearest'));
set(gca,'xtick',[],'ytick',[],'LineWidth',2,'XColor',[1 0 0],'YColor',[1 0 0]);
colormap(gray(256));box on;%axis off;

figure,imagesc(IMql);colormap(gray(256));axis off;hold on;
rectangle('Position',[TLx1,TLy1,Kx,Ky],'EdgeColor', 'r','LineWidth', 1.5,'LineStyle','-');
axes('position',pos1)
imagesc(IMql(TLy1:TLy1+Ky,TLx1:TLx1+Kx,:));colormap(gray(256));box on;
set(gca,'xtick',[],'ytick',[],'LineWidth',3,'XColor',[1 0 0],'YColor',[1 0 0]);

figure,imagesc(IMqh);colormap(gray(256));axis off;hold on;
rectangle('Position',[TLx1,TLy1,Kx,Ky],'EdgeColor', 'r','LineWidth', 1.5,'LineStyle','-');
axes('position',pos1)
imagesc(IMqh(TLy1:TLy1+Ky,TLx1:TLx1+Kx,:));colormap(gray(256));box on;
set(gca,'xtick',[],'ytick',[],'LineWidth',3,'XColor',[1 0 0],'YColor',[1 0 0]);

figure,imagesc(IMqs);colormap(gray(256));axis off;hold on;
rectangle('Position',[TLx1,TLy1,Kx,Ky],'EdgeColor', 'r','LineWidth', 1.5,'LineStyle','-');
axes('position',pos1)
imagesc(IMqs(TLy1:TLy1+Ky,TLx1:TLx1+Kx,:));colormap(gray(256));box on;
set(gca,'xtick',[],'ytick',[],'LineWidth',3,'XColor',[1 0 0],'YColor',[1 0 0]);

figure,imagesc(IM);colormap(gray(256));axis off;hold on;
rectangle('Position',[TLx1,TLy1,Kx,Ky],'EdgeColor', 'r','LineWidth', 1.5,'LineStyle','-');
axes('position',pos1)
imagesc(IM(TLy1:TLy1+Ky,TLx1:TLx1+Kx,:));colormap(gray(256));box on;
set(gca,'xtick',[],'ytick',[],'LineWidth',3,'XColor',[1 0 0],'YColor',[1 0 0]);