%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Demo of Convergence of Threshold Adaptation Techniques
%
% O. A. Elgendy and S. H. Chan,
% "Optimal Threshold Design for Quanta Image Sensor," 
% IEEE Transactions on Computational Imaging, 2017.
% http://ieeexplore.ieee.org/document/8169093/
%
% Omar Elgendy and Stanley Chan
% Copyright 2017
% Purdue University, West Lafayette, In, USA.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

clc,clear,close all
rng('default');

% Load an Image
addpath(genpath('./SSIP_ImageData/'));
image                         =  dir('./SSIP_ImageData/*.JPG');
NbImages                      =  length(image);

% QIS Simulation Parameters
K                             =  5;         % Overasampling Factor
qmax                          =  16;        % Maximum Threshold
T_adapt                       =  10;        % Number of frames for threshold adaptation
NbRndSamples                  =  50;        % Number of random samples

% Parameters of Bisection Technique
gamma1                        =  0.5;       % Target proportion of ones
Qblocksize                    =  1;         % 1 Threshold shared by KxK jots

% Parameters of Markov Chain Technique
LSet                          =  [4 3 2];
Betaset                       =  0.25;
[x1,x2]                       =  meshgrid(LSet,Betaset);
LBeta                         =  [x1(:) x2(:)];
NbCombLBeta                   =  length(LBeta);

fprintf('------------------------------------------------------------------\n')
fprintf('--------------Monte Carlo Simulations for 1 image-----------------\n')
fprintf('------------------------------------------------------------------\n')
i                             =  3;
IM                            =  im2double(imread(image(i).name));
IM                            =  imresize(IM,0.1);
IM                            =  rgb2gray(IM);
[rows,cols]                   =  size(IM);
g                             =  ones(K);
c                             =  min(IM,1);
alpha                         =  qmax-1;
s                             =  max(min(kron(c,g),1),0);
% Oracel Threshold Map
oracleQmap                    =  floor(alpha*c)+1;
oracleQmap                    =  max(min(oracleQmap,qmax),1);

% Find pixels with low, center, and high thresholds
x1=find(oracleQmap<= 0.2*qmax);x1=x1(randi(length(x1),33,1));            % Indices of low thresholds
x2=find(abs(oracleQmap-qmax/2)<=0.1);x2=x2(randi(length(x2),33,1));      % Indices of center thresholds
x3=find(oracleQmap>= 0.9*qmax);x3=x3(randi(length(x1),33,1));            % Indices of high thresholds
PixIdx=zeros(rows,cols);
PixIdx(x1)=1;
PixIdx(x2)=1;
PixIdx(x3)=1;
PixIdx=logical(PixIdx);
NbPixels                      =  1*sum(PixIdx(:));
GTPixels                      =  oracleQmap(PixIdx);

MCqiter                       =  zeros(T_adapt+1,  NbPixels,NbCombLBeta,NbRndSamples);
MCMSE                         =  zeros(T_adapt+1,  NbCombLBeta,NbRndSamples);
[Bqiter_l,Bqiter_c,Bqiter_h]  =  deal(zeros(T_adapt+1-3,NbPixels,NbRndSamples));
BMSE                          =  zeros(T_adapt+1-3,NbRndSamples);
parfor r=1:NbRndSamples
    
    fprintf('Random Sample No. %d\n',r);
    
    y=zeros(rows*K,cols*K,T_adapt);
    for w=1:T_adapt+1
        y(:,:,w)=poissrnd(alpha*s); % Photon Count
    end
    
    [BQmap,~,~,~,err,iter_B,~] =  BisectionAdapt(y,qmax,Qblocksize,oracleQmap,T_adapt,PixIdx,1);
    Bqiter_l(:,:,r)            =  squeeze(iter_B(:,1,:))';
    Bqiter_c(:,:,r)            =  squeeze(iter_B(:,2,:))';
    Bqiter_h(:,:,r)            =  squeeze(iter_B(:,3,:))';
    BMSE(:,r)                  =  err';
    
    for k=1:NbCombLBeta
        [MCQmap,~,~,err,qiter] =  MCAdapt(y,qmax,oracleQmap,LBeta(k,1),LBeta(k,2),PixIdx);
        MCqiter(:,:,k,r)       =  qiter';
        MCMSE(:,k,r)           =  err';
    end
    
end
MCqiter_av                     =  mean(MCqiter,4);
Bqiter_c_av                    =  mean(Bqiter_c,3);
MCMSE_av                       =  mean(MCMSE,3);
BMSE_av                        =  mean(BMSE,2);

fprintf('------------------------------------------------------------------\n')
fprintf('-----------Monte Carlo Simulations for 100 images-----------------\n')
fprintf('------------------------------------------------------------------\n')

MCMSE_tstimages                =  zeros(T_adapt+1,  NbCombLBeta,NbImages,NbRndSamples);
BMSE_tstimages                 =  zeros(T_adapt+1-3,NbImages,NbRndSamples);

parfor i=1:NbImages
    
    fprintf('Image %d\n',i);
    IM                         =  im2double(imread(image(i).name));
    IM                         =  imresize(IM,0.1);
    IM                         =  rgb2gray(IM);
    [rows,cols]                =  size(IM);
    g                          =  ones(K);
    c                          =  min(IM,1);
    alpha                      =  qmax-1;
    oracleQmap                 =  floor(alpha*c)+1;
    oracleQmap                 =  max(min(oracleQmap,qmax),1);
    s                          =  max(min(kron(c,g),1),0);
    PixIdx                     =  [];
    
    for r=1:NbRndSamples
        
        fprintf('Image %d, Random Sample No. %d\n',i,r);
        y=zeros(rows*K,cols*K,T_adapt+1);
        for w=1:T_adapt+1
            y(:,:,w)=poissrnd(alpha*s); % Photon Count
        end
        
        [BQmap,~,~,~,err,~,~]        =  BisectionAdapt(y,qmax,Qblocksize,g,oracleQmap,T_adapt,PixIdx,1);
        BMSE_tstimages(:,i,r)        =  err';
        
        for k=1:NbCombLBeta
            [MCQmap,~,~,err,~]       =  MCAdapt(y,qmax,oracleQmap,LBeta(k,1),LBeta(k,2),PixIdx);
            MCMSE_tstimages(:,k,i,r) =  err';
        end
    end
end
MCMSE_tstimages_av                   =  mean(mean(MCMSE_tstimages,4),3);
BMSE_tstimages_av                    =  mean(mean(BMSE_tstimages,3),2);

%%
figure('Color',[1 1 1],'Position',[78 277 909 481])
set(gca,'fontsize',18,'fontweight','bold','TickLabelInterpreter','latex')
xlim([1 11])
xlabel('Major Iteration');ylabel('Mean Square Error')
set(groot, 'defaultAxesTickLabelInterpreter','latex'); set(groot, 'defaultLegendInterpreter','latex');
hold on;
plot(MCMSE_tstimages_av(:,1),'-bx','linewidth',3,'markersize',12)
plot(MCMSE_tstimages_av(:,2),'-bs','linewidth',3,'markersize',12)
plot(MCMSE_tstimages_av(:,3),'-b^','linewidth',3,'markersize',12)
plot([BMSE_tstimages_av; BMSE_tstimages_av(end)*ones(4,1)],'-ro','linewidth',3,'markersize',12)
legend('[18], $L=4$','[18], $L=3$','[18], $L=2$','Proposed',10,'location','best')%beta =0.25
box off


%%
figure;
subplot(311);l=10;
hold on;
plot(MCqiter_av(:,l,2),'-bx','linewidth',3,'markersize',12)
plot(MCqiter_av(:,l,6),'-bs','linewidth',3,'markersize',12)
plot(MCqiter_av(:,l,10),'-b^','linewidth',3,'markersize',12)
plot([Bqiter_c_av(:,l);Bqiter_c_av(end,l)*ones(4,1)],'-ro','linewidth',3,'markersize',12)
line([1 T_adapt+3],[GTPixels(l) GTPixels(l)],'color',[0 0 0],'linewidth',2,'linestyle','--')
set(gca,'fontsize',18,'fontweight','bold','TickLabelInterpreter','latex')
xlabel('Major Iteration');ylabel('Threshold')
set(groot, 'defaultAxesTickLabelInterpreter','latex'); set(groot, 'defaultLegendInterpreter','latex');
legend_str=[{'[18], $L=4$'},{'[18], $L=3$'},{'[18], $L=2$'},{'Proposed'},{'$q^*(c)$'}];
h=columnlegend(5, legend_str, 'Location', 'best');
xlim([1 11]);ylim([8 16])

subplot(312);l=38;
hold on;
plot(MCqiter_av(:,l,2),'-bx','linewidth',3,'markersize',12)
plot(MCqiter_av(:,l,6),'-bs','linewidth',3,'markersize',12)
plot(MCqiter_av(:,l,10),'-b^','linewidth',3,'markersize',12)
plot([Bqiter_c_av(:,l);Bqiter_c_av(end,l)*ones(4,1)],'-ro','linewidth',3,'markersize',12)
line([1 T_adapt+3],[GTPixels(l) GTPixels(l)],'color',[0 0 0],'linewidth',2,'linestyle','--')
set(gca,'fontsize',18,'fontweight','bold','TickLabelInterpreter','latex')
xlabel('Major Iteration');ylabel('Threshold')
set(groot, 'defaultAxesTickLabelInterpreter','latex'); set(groot, 'defaultLegendInterpreter','latex');
xlim([1 11]);ylim([5.5 9])

subplot(313);l=9;
hold on;
plot(MCqiter_av(:,l,2),'-bx','linewidth',3,'markersize',12)
plot(MCqiter_av(:,l,6),'-bs','linewidth',3,'markersize',12)
plot(MCqiter_av(:,l,10),'-b^','linewidth',3,'markersize',12)
plot([Bqiter_c_av(:,l);Bqiter_c_av(end,l)*ones(4,1)],'-ro','linewidth',3,'markersize',12)
line([1 T_adapt+3],[GTPixels(l) GTPixels(l)],'color',[0 0 0],'linewidth',2,'linestyle','--')
set(gca,'fontsize',18,'fontweight','bold','TickLabelInterpreter','latex')
xlabel('Major Iteration');ylabel('Threshold')
set(groot, 'defaultAxesTickLabelInterpreter','latex'); set(groot, 'defaultLegendInterpreter','latex');
xlim([1 11]);ylim([1 8])