%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Demo of Reconstruction Perfomance of different Threshold Maps
%
% O. A. Elgendy and S. H. Chan,
% "Optimal Threshold Design for Quanta Image Sensor,"
% IEEE Transactions on Computational Imaging, 2017.
% http://ieeexplore.ieee.org/document/8169093/
%
% Omar Elgendy and Stanley Chan
% Copyright 2017
% Purdue University, West Lafayette, In, USA.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

clc,clear,close all
rng('default');
t1                           =  tic;
addpath(genpath('./ImageData/'));
addpath(genpath('./Utilities'));
image                        =  dir('./ImageData/*.JPG');
NbImages                     =  length(image);

% QIS Simulation Parameters
K                            =  4;                % Spatial  Overasampling Factor
T_adapt                      =  8;				  % Frames for Threshold Adaptation
T_recon                      =  7;                % Frames for Image Recosntruction
T                            =  T_adapt+T_recon;  % Temporal Overasampling Factor
Qsize                        =  [1 K 2*K];        % Kernel size of pixels that share the same threshold
qmax                         =  16;               % Maximum Threshold
NbRndSamples                 =  50;               % Number of Random Samples for MonteCarlo
QmapU                        =  [1 5 10 16];      % Uniform Threshold
g                            =  ones(K)/K^2;      % Interpolation kernel
NbBlocks                     =  length(Qsize);
NbqU                         =  length(QmapU);

PSNR_qs                      =  zeros(NbImages,NbBlocks,NbRndSamples);
PSNR_qu                      =  zeros(NbImages,NbqU,NbRndSamples);
PSNR_CR_desQ_descDur         =  zeros(NbImages,NbRndSamples);
PSNR_CR_ascQ_descDur         =  zeros(NbImages,NbRndSamples);



for i=1:NbImages
    
    fprintf('Image No.%d\n',i);
    IM = imresize(im2double(imread(image(i).name)),0.1);
    if size(IM,3)~=1;IM      =  rgb2gray(IM);end
    [rows,cols]              =  size(IM);
    alpha                    =  K^2*(qmax-1);    % Sensor gain
    oracleQmap               =  floor(alpha*IM/K^2)+1;
    
    for r=1:NbRndSamples
        
        fprintf('Image No. %d, Random Sample No.%d\n',i,r);
        
        y                    =  generatePhotonCounts(IM,K,T,alpha,g);
        
        % option variable for the "generateQmap" function
        opts.y               =  y;
        opts.qmax            =  qmax;
        opts.oracleQmap      =  oracleQmap;
        opts.T_adapt         =  T_adapt;
                
        temp=zeros(NbBlocks,1);
        for j=1:NbBlocks
            opts.Qblock      =  Qsize(j);
            [QmapLR,QmapHR]  =  generateQmap([rows,cols],K,'bisection',opts);
            reconFrames      =  T_adapt + (1:T_recon);
            b                =  1*(y(:,:,reconFrames)>=repmat(QmapHR,1,1,T_recon));
            IM_qs            =  imageReconstruct(b,K,alpha,QmapLR);
            temp(j,1)        =  psnr(IM_qs,IM);
            %fprintf(' %dx%d: %2.0fdB \n',Qblock,Qblock,PSNR_qs(i,j,r));
        end
        PSNR_qs(i,:,r)=temp;
        
        temp=zeros(NbqU,1);
        for j=1:NbqU
            opts.q           =  QmapU(j);
            [QmapLR,QmapHR]  =  generateQmap([rows,cols],K,'uniform',opts);
            b                =  1*(y>=repmat(QmapHR,1,1,T));
            IM_qu            =  imageReconstruct(b,K,alpha,QmapLR);
            temp(j,1)        =  psnr(IM_qu,IM);
            %fprintf(' q=%d: %2.0fdB \n',QmapU(j),PSNR_qu(i,j,r));
        end
        PSNR_qu(i,:,r)=temp;
        
        TCR                         =  K^2*T;         %Temporal Overasampling Factor
        sampling_interval_Dur       =  logspace(0,-log10(TCR),TCR);
        y0CR=zeros(rows,cols,TCR);
        for t=1:TCR
            y0CR(:,:,t)=poissrnd(alpha*IM*sampling_interval_Dur(t)/K^2); % Photon Count
        end
        
        Q_sequence                  =  fliplr(1:qmax)';
        [IM_CR_descQ_descDur,PSNR_CR_desQ_descDur(i,r)]    =  ConditionalResetAlg(y0CR,TCR,Q_sequence,IM);
        %fprintf('Descending Q: PSNR=%2.2f\n',PSNR_CR_desQ_descDur(i,r));
        
        Q_sequence                  =  (1:qmax)';
        [IM_CR_ascQ_descDur,PSNR_CR_ascQ_descDur(i,r)]     =  ConditionalResetAlg(y0CR,TCR,Q_sequence,IM);
        %fprintf(' Ascending Q: PSNR=%2.2f\n',PSNR_CR_ascQ_descDur(i,r));
    end
end
fprintf('CR Ascending Q sequence              &%2.2f &%2.2f\n',mean(mean(PSNR_CR_ascQ_descDur)),mean(sqrt(var(PSNR_CR_ascQ_descDur,[],2))))
fprintf('CR Descending Q sequence             &%2.2f &%2.2f\n',mean(mean(PSNR_CR_desQ_descDur)),mean(sqrt(var(PSNR_CR_desQ_descDur,[],2))))
fprintf('Uniform Q-Map q=1:                   &%2.2f &%2.2f\n',mean(mean(PSNR_qu(:,1,:))),mean(sqrt(var(PSNR_qu(:,1,:),[],3))))
fprintf('Uniform Q-Map q=10:                  &%2.2f &%2.2f\n',mean(mean(PSNR_qu(:,3,:))),mean(sqrt(var(PSNR_qu(:,3,:),[],3))))
fprintf('Uniform Q-Map q=5:                   &%2.2f &%2.2f\n',mean(mean(PSNR_qu(:,2,:))),mean(sqrt(var(PSNR_qu(:,2,:),[],3))))
fprintf('Adaptive 2Kx2K Q-Map:                &%2.2f &%2.2f\n',mean(mean(PSNR_qs(:,3,:))),mean(sqrt(var(PSNR_qs(:,3,:),[],3))))
fprintf('Adaptive KxK Q-Map:                  &%2.2f &%2.2f\n',mean(mean(PSNR_qs(:,2,:))),mean(sqrt(var(PSNR_qs(:,2,:),[],3))))
fprintf('Adaptive 1x1 Q-Map:                  &%2.2f &%2.2f\n',mean(mean(PSNR_qs(:,1,:))),mean(sqrt(var(PSNR_qs(:,1,:),[],3))))
totaltime=toc(t1);
