%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Demo of HDR image Reconstruction using QIS
%
% O. A. Elgendy and S. H. Chan,
% "Optimal Threshold Design for Quanta Image Sensor,"
% IEEE Transactions on Computational Imaging, 2017.
% http://ieeexplore.ieee.org/document/8169093/
%
% Omar Elgendy and Stanley Chan
% Copyright 2017
% Purdue University, West Lafayette, In, USA.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

clc,clear,close all
rng('default');

addpath(genpath('./Utilities/HDRToolbox/source_code/'))  % HDR Toolbox
addpath(genpath('./ImageData/HDREyeDataset/C18'))        % HDREye Dataset

image                =  dir('./ImageData/HDREyeDataset/C18/*.JPG');
NbExposure           =  length(image);
K                    =  4;                %  Spatial  Overasampling Factor
T_adapt              =  8;                %  Frames for Threshold Adaptation
T_recon              =  5;                %  Frames for Image Recosntruction
T                    =  T_adapt+T_recon;  %  Temporal Overasampling Factor
Qsize                =  K;                %  KxK pixels share the same threshold
qmax                 =  16;               %  Maximum Quantization Theshold
gamma1               =  0.5;              %  Target proportion of ones
scale                =  0.05;             %  Downscale the original image
g                    =  ones(K)/K^2;      %  Interpolation kernel
alpha                =  K^2*(qmax-1);     %  Sensor gain

% option variable for the "generateQmap" function
opts.qmax            =  qmax;
opts.T_adapt         =  T_adapt;

RGB                  =  im2double(imread(image(1).name));
RGB                  =  imresize(RGB,scale);
[rows,cols,~]        =  size(RGB);
[IMGT,IMql,IMqh,IMqchess,IMqpattern,IMqs,IMqs_1x1]         =  deal(zeros(rows,cols,1,NbExposure));
[PSNRq1,PSNRqh,PSNRqs,PSNRqs_1x1,PSNRqchess,PSNRqpattern]  =  deal(zeros(NbExposure,1));
fprintf('Processing on %d Exposures\n',NbExposure)
for i=1:NbExposure
    
    fprintf('Exposure%d:',i);
    RGB                  =  im2double(imread(image(i).name));
    RGB                  =  imresize(RGB,scale);
    YCBCR                =  rgb2ycbcr(RGB);
    IM                   =  YCBCR(:,:,1);
    figure;imshow(IM);
    %-----------------Ground Truth-------------------------------------
    c                    =  max(0,IM);
    oracleQmap           =  min(floor(alpha*c/K^2)+1,qmax);
    opts.oracleQmap      =  oracleQmap;
    y                    =  generatePhotonCounts(c,K,T,alpha,g);
    IMGT(:,:,:,i)        =  IM;
    %---------------Uniform Qmap q=1-----------------------------------
    opts.q               =  1;
    [QmapLR,QmapHR]      =  generateQmap([rows,cols],K,'uniform',opts);
    b                    =  double(y>=repmat(QmapHR,1,1,T));
    IMql(:,:,1,i)        =  imageReconstruct(b,K,alpha,QmapLR);
    PSNRq1(i)            =  psnr(IMql(:,:,1,i),IM);
    fprintf(' PSNR_q1 %2.0f dB /',PSNRq1(i));
    %---------------Uniform Qmap q=qmax---------------------------------
    opts.q               =  qmax;
    [QmapLR,QmapHR]      =  generateQmap([rows,cols],K,'uniform',opts);
    b                    =  double(y>=repmat(QmapHR,1,1,T));
    IMqh(:,:,1,i)        =  imageReconstruct(b,K,alpha,QmapLR);
    PSNRqh(i)            =  psnr(IMqh(:,:,1,i),IM);
    fprintf(' PSNR_qh %2.0f dB /',PSNRqh(i));
    %---------------Adapted KxK Q-map-------------------------------
    opts.y               =  y;
    opts.Qblock          =  K;
    [QmapLR,QmapHR]      =  generateQmap([rows,cols],K,'bisection',opts);
    framesqs             =  T_adapt + (1:T_recon);
    b                    =  double(y(:,:,framesqs)>=repmat(QmapHR,1,1,T_recon));
    IMqs(:,:,1,i)        =  imageReconstruct(b,K,alpha,QmapLR);
    PSNRqs(i)            =  psnr(IMqs(:,:,1,i),IM);
    fprintf(' PSNR_qs %2.0f dB /',PSNRqs(i));
    %---------------Adapted 1x1 Q-map-------------------------------
    opts.Qblock          =  1;
    [QmapLR,QmapHR]      =  generateQmap([rows,cols],K,'bisection',opts);
    b                    =  double(y(:,:,framesqs)>=repmat(QmapHR,1,1,T_recon));
    IMqs_1x1(:,:,1,i)    =  imageReconstruct(b,K,alpha,QmapLR);
    PSNRqs_1x1(i)        =  psnr(IMqs_1x1(:,:,1,i),IM);
    fprintf(' PSNR_qs_1x1 %2.0f dB \n',PSNRqs_1x1(i));
end

%% Ramans Image Fusion Tonemapping (From HDR_Toolbox)
RamansGT        = RamanTMO([],'', '',IMGT);
Ramansql        = RamanTMO([],'', '',IMql);
Ramansqh        = RamanTMO([],'', '',IMqh);
Ramansqs        = RamanTMO([],'', '',IMqs);
Ramansqs_1x1    = RamanTMO([],'', '',IMqs_1x1);

%% Mertens Image Fusion Tonemapping (From HDR_Toolbox)
MertensGT       = MertensTMO([], '', '', IMGT, 1, 1, 1,0);
Mertensql       = MertensTMO([], '', '', IMql, 1, 1, 1,0);
Mertensqh       = MertensTMO([], '', '', IMqh, 1, 1, 1,0);
Mertensqs       = MertensTMO([], '', '', IMqs, 1, 1, 1,0);
Mertensqs_1x1   = MertensTMO([], '', '', IMqs_1x1, 1, 1, 1,0);
%----------------------------
%----------------------------
fprintf('PSNR ML-CF Raman q1        =  %2.2f dB\n',psnr(Ramansql,RamansGT));
fprintf('PSNR ML-CF Raman qh        =  %2.2f dB\n',psnr(Ramansqh,RamansGT));
fprintf('PSNR ML-CF Qmap Raman      =  %2.2f dB\n',psnr(Ramansqs,RamansGT));
fprintf('---------------------------------------\n');
fprintf('PSNR ML-CF Mertens q1      =  %2.2f dB\n',psnr(Mertensql,MertensGT));
fprintf('PSNR ML-CF Mertens qh      =  %2.2f dB\n',psnr(Mertensqh,MertensGT));
fprintf('PSNR ML-CF Qmap Mertens    =  %2.2f dB\n',psnr(Mertensqs,MertensGT));
figure;imshow(Mertensql)
figure;imshow(Mertensqh)
figure;imshow(Mertensqs)

%%
close all
% % c31
% TLx1=115;TLy1=600;
% Kx=40;Ky=40;
% TLx2=375;TLy2=110;
% position1 = [0.61 0.11 0.3 0.3];
% position2 = [0.61 0.63 0.3 0.3];
%c18
TLx1=250;TLy1=125;
Kx=40;Ky=40;
TLx2=730;TLy2=580;
position1 = [0.61 0.63 0.3 0.3];
position2 = [0.13 0.11 0.3 0.3];
% % c20
% TLx1=50;TLy1=320;
% Kx=45;Ky=45;
% TLx2=675;TLy2=360;
% position1 = [0.61 0.11 0.3 0.3];
% position2 = [0.61 0.63 0.3 0.3];

scale=[min(Mertensql(:)) max(Mertensql(:))];
figure,imagesc(Mertensql);colormap(gray(256));axis off;hold on;
rectangle('Position',[TLx1,TLy1,Kx,Ky],'EdgeColor', 'r','LineWidth', 2,'LineStyle','-');
rectangle('Position',[TLx2,TLy2,Kx,Ky],'EdgeColor', 'b','LineWidth', 2,'LineStyle','-');
axes('position',position1)
imagesc(Mertensql(TLy1:TLy1+Ky,TLx1:TLx1+Kx),scale);colormap(gray(256));box on;
set(gca,'xtick',[],'ytick',[],'LineWidth',3,'XColor',[1 0 0],'YColor',[1 0 0]);
axes('position',position2)
imagesc(Mertensql(TLy2:TLy2+Ky,TLx2:TLx2+Kx),scale);colormap(gray(256));box on;
set(gca,'xtick',[],'ytick',[],'LineWidth',3,'XColor',[0 0 1],'YColor',[0 0 1]);

scale=[min(Mertensqh(:)) max(Mertensqh(:))];
figure,imagesc(Mertensqh);colormap(gray(256));axis off;hold on;
rectangle('Position',[TLx1,TLy1,Kx,Ky],'EdgeColor', 'r','LineWidth', 2,'LineStyle','-');
rectangle('Position',[TLx2,TLy2,Kx,Ky],'EdgeColor', 'b','LineWidth', 2,'LineStyle','-');
axes('position',position1)
imagesc(Mertensqh(TLy1:TLy1+Ky,TLx1:TLx1+Kx),scale);colormap(gray(256));box on;
set(gca,'xtick',[],'ytick',[],'LineWidth',3,'XColor',[1 0 0],'YColor',[1 0 0]);
axes('position',position2)
imagesc(Mertensqh(TLy2:TLy2+Ky,TLx2:TLx2+Kx),scale);colormap(gray(256));box on;
set(gca,'xtick',[],'ytick',[],'LineWidth',3,'XColor',[0 0 1],'YColor',[0 0 1]);

scale=[min(Mertensqs(:)) max(Mertensqs(:))];
figure,imagesc(Mertensqs);colormap(gray(256));axis off;hold on;
rectangle('Position',[TLx1,TLy1,Kx,Ky],'EdgeColor', 'r','LineWidth', 2,'LineStyle','-');
rectangle('Position',[TLx2,TLy2,Kx,Ky],'EdgeColor', 'b','LineWidth', 2,'LineStyle','-');
axes('position',position1)
imagesc(Mertensqs(TLy1:TLy1+Ky,TLx1:TLx1+Kx),scale);colormap(gray(256));box on;
set(gca,'xtick',[],'ytick',[],'LineWidth',3,'XColor',[1 0 0],'YColor',[1 0 0]);
axes('position',position2)
imagesc(Mertensqs(TLy2:TLy2+Ky,TLx2:TLx2+Kx),scale);colormap(gray(256));box on;
set(gca,'xtick',[],'ytick',[],'LineWidth',3,'XColor',[0 0 1],'YColor',[0 0 1]);

scale=[min(MertensGT(:)) max(MertensGT(:))];
figure,imagesc(MertensGT);colormap(gray(256));axis off;hold on;
rectangle('Position',[TLx1,TLy1,Kx,Ky],'EdgeColor', 'r','LineWidth', 2,'LineStyle','-');
rectangle('Position',[TLx2,TLy2,Kx,Ky],'EdgeColor', 'b','LineWidth', 2,'LineStyle','-');
axes('position',position1)
imagesc(MertensGT(TLy1:TLy1+Ky,TLx1:TLx1+Kx),scale);colormap(gray(256));box on;
set(gca,'xtick',[],'ytick',[],'LineWidth',3,'XColor',[1 0 0],'YColor',[1 0 0]);
axes('position',position2)
imagesc(MertensGT(TLy2:TLy2+Ky,TLx2:TLx2+Kx),scale);colormap(gray(256));box on;
set(gca,'xtick',[],'ytick',[],'LineWidth',3,'XColor',[0 0 1],'YColor',[0 0 1]);