%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Monte Carlo Simulation for showing the QIS Phase Transition Behavior 
%
% O. A. Elgendy and S. H. Chan,
% "Optimal Threshold Design for Quanta Image Sensor," 
% IEEE Transactions on Computational Imaging, 2017.
% http://ieeexplore.ieee.org/document/8169093/
% 
% Omar Elgendy and Stanley Chan
% Copyright 2017
% Purdue University, West Lafayette, In, USA.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

clc,clear,close all
rng('default')

%% QIS Simulation Parameters
c0         =  0.5;   % Ground Truth
alpha      =  300;   % Sensor Gain
RndSamples =  5000;   % Number of Random Samples for Monte Carlo
% Set of Spatial Oversampling Factors
Ksqset     =  (2:10).^2;
Kcar       =  length(Ksqset);
% Set of Temporal Oversampling Factors
Tset       =  [1 10:10:100];
Tcar       =  length(Tset);
% Set of Quantization Thresholds
Qset       =  1:200;
Qcar       =  length(Qset);

% Organize all combinations for parallel processing
j                       =  1;
NbParam                 =  Kcar*Qcar*Tcar;
param                   =  zeros(NbParam,3);
for q                   =  Qset
    for Ksq             =  Ksqset
        for T           =  Tset
            param(j,:)  =  [q Ksq T];
            j           =  j+1;
        end
    end
end

%% Main Loop
[SNR_all,chat_all,gamma1_all] = deal(zeros(NbParam,RndSamples));
for j=1:NbParam
    q                   =  param(j,1);
    Ksq                 =  param(j,2);
    T                   =  param(j,3);
    fprintf('q = %d, K^2 = %d, T = %d \n', q, Ksq, T);
    s                   =  alpha*c0*ones(sqrt(Ksq))/Ksq;
    parfor i=1:RndSamples
        Y               =  poissrnd(repmat(s,1,1,T));
        B               =  1*(Y>=q);
        gamma1_all(j,i) =  mean(B(:));
        chat_all(j,i)   =  Ksq/alpha*gammaincinv(1-gamma1_all(j,i), q,'upper');
        SNR_all(j,i)    =  10*log10(c0^2/(chat_all(j,i)-c0)^2);
    end
end

%% Re-Organize Data
chat_all(chat_all==inf)  = 1;
gamma1      =  mean(gamma1_all,2);
chat        =  mean(chat_all,2);
snr         =  mean(SNR_all,2);
mcsnr       =  10*log10(c0^2./mean((chat_all-c0).^2,2));
mcmse       =  10*log10(mean((chat_all-c0).^2,2));
gamma1_std  =  sqrt(var(gamma1_all,[],2));
chat_std    =  sqrt(var(chat_all,[],2));
snr_std     =  sqrt(var(SNR_all,[],2));
mcsnr_std   =  sqrt(var((chat_all-c0).^2,[],2));

[eta,eta_std,Gamma1,Gamma1_std,SNR,SNR_std,mcSNR,mcSNR_std,mcMSE] = deal(zeros(Qcar,Kcar,Tcar));
for j=1:NbParam
    a                =  param(j,1);
    b                =  find(Ksqset==param(j,2));
    c                =  find(Tset==param(j,3));
    eta(a,b,c)       =  chat(j)/c0;
    eta_std(a,b,c)   =  chat_std(j);
    Gamma1(a,b,c)    =  gamma1(j);
    Gamma1_std(a,b,c)=  gamma1_std(j);
    SNR(a,b,c)       =  snr(j);
    SNR_std(a,b,c)   =  snr_std(j);
    mcSNR(a,b,c)     =  mcsnr(j);
    mcSNR_std(a,b,c) =  mcsnr_std(j);
    mcMSE(a,b,c)     =  mcmse(j);
end

%% Draw Phase Transition 
qmax      =  100;      % Maximum Quantization Threshold                     
peak      =  3;
kidx      =  1;        % Index of Spatial Oversampling Factor in the set
Tidx      =  6;		   % Index of Temporal Oversampling Factor in the set
KT        =  Ksqset(kidx)*Tset(Tidx);
theta     =  alpha*c0/Ksqset(kidx); % Light Exposure Value
qstar     =  floor(theta)+1;        % Oracle Threshold

mcGamma   =  squeeze(Gamma1(:,kidx,Tidx)); % Monte Carlo Gamma
mcETA     =  eta(:,kidx,Tidx);             % Monte Carlo Eta

Psi       =  real(gammainc(theta,Qset,'upper'));
p0        =  Psi.^KT;
p1        =  (1-Psi).^KT;
pSuccess  =  1 - p0 - p1;
ETA       =  pSuccess + p1/c0;                  % Exact Eta (Calculated using probability distriction)
MSE       =  10*log10(c^2 * p1 + (1-c)^2 * p0); % Exact MSE (Calculated using probability distriction)
SNR       =  -10*log10(p1+p0*(1-1/c)^2);        % Exact SNR (Calculated using probability distriction)  

delta     =  2*1e-4;
Epsilon   =  1-(delta/2)^(1/KT);
qrange    =  Qset(Psi>=Epsilon & Psi<=1-Epsilon);
ql        =  qrange(1);
qh        =  qrange(end);

qlsnr     =  35.8;
qhsnr     =  41.75;
bdl       =  0.264;
bdh       =  0.63;

figure('Color',[1 1 1],'Position',[7 -223 1108 908])
subplot(211)
set(groot, 'defaultAxesTickLabelInterpreter','latex'); set(groot, 'defaultLegendInterpreter','latex');
h1 = patch([0 ql ql 0],[0 0 peak peak],[1 0 0]);set(h1,'FaceAlpha',0.2,'LineStyle','none');hold on
h2 = patch([ql qh qh ql],[0 0 peak peak],[0 1 0]);set(h2,'FaceAlpha',0.2,'LineStyle','none');
h3 = patch([qh qmax qmax qh],[0 0 peak peak],[1 0 0]);set(h3,'FaceAlpha',0.2,'LineStyle','none');
h4 = plot(Qset,ETA,'Color',[0.25 0.5 0.25],'linewidth',1.5);
h5 = plot(Qset,mcETA,'--b','linewidth',3);
h6 = plot(Qset,1-Psi,'-k','linewidth',1.5);
h7 = plot(mcGamma,'--k','linewidth',3);
xlim([Qset(1) qmax]);ylim([0 1.2])
xlabel('$\textrm{Threshold } q$')
h  = legend([h2 h3 h4 h5 h6 h7],{'$q\in\mathcal{Q}_\theta$','$q\notin\mathcal{Q}_\theta$',...
     'Exact $E[\,\widehat{c}\,]/c$','Monte-Carlo $E[\,\widehat{c}\,]/c$','Exact $1-E[\gamma_q(c)]$','Monte-Carlo $1-E[\gamma_q(c)]$'}); 
set(h,'Interpreter','latex')
set(gca,'fontsize',20,'fontweight','bold','LineWidth',2)
gamma_star =  0.4891;
lwidth     =  1.5;
c1         =  [0.75 0.25 0.25];
line([Qset(1) qlsnr],[bdh bdh],'linestyle','-.','linewidth',lwidth,'Color',c1);
line([Qset(1) qstar],[gamma_star gamma_star],'linestyle','-.','linewidth',lwidth,'Color',c1);
line([Qset(1) qhsnr],[bdl bdl],'linestyle','-.','linewidth',lwidth,'Color',c1);
line([qlsnr qlsnr],[bdh 0],'linestyle','-.','linewidth',lwidth,'Color',c1);
line([qstar qstar],[gamma_star 0],'linestyle','-.','linewidth',lwidth,'Color',c1);
line([qhsnr qhsnr],[bdl 0],'linestyle','-.','linewidth',lwidth,'Color',c1);

subplot(212)
h1 = patch([0 ql ql 0],[0 0 40 40],[1 0 0]);set(h1,'FaceAlpha',0.2,'LineStyle','none');hold on;
h2 = patch([ql qh qh ql],[0 0 40 40],[0 1 0]);set(h2,'FaceAlpha',0.2,'LineStyle','none');
h3 = patch([qh qmax qmax qh],[0 0 40 40],[1 0 0]);set(h3,'FaceAlpha',0.2,'LineStyle','none');
line([qlsnr qlsnr],[0 40],'linestyle','-.','linewidth',lwidth,'Color',c1);
line([qstar qstar],[0 40],'linestyle','-.','linewidth',lwidth,'Color',c1);
line([qhsnr qhsnr],[0 40],'linestyle','-.','linewidth',lwidth,'Color',c1);
semilogy(Qset,mcSNR(:,kidx,Tidx),'LineWidth',3,'Color',[0 0 1]);
xlabel('$\textrm{Threshold } q$')
ylabel('SNR in dB') % left y-axis
set(gca,'fontsize',20,'fontweight','bold','LineWidth',2,'Box','off','YColor',[0 0 0],'XLim',[Qset(1) qmax],'YLim',[0 40]);

%% Average Bit Density
KT                    =  200;
theta                 =  linspace(0,50,10000);
qstar                 =  floor(theta)+1;
qstarround            =  round(theta) ;
avBitDensitymyqs      =  gammainc(theta,qstar,'lower');
stdBitDensitymyqs     =  1/KT*gammainc(theta,qstar,'lower').*gammainc(theta,qstar,'upper');
avBitDensityqsround   =  gammainc(theta,qstarround,'lower');
stdBitDensityqsround  =  1/KT*gammainc(theta,qstarround,'lower').*gammainc(theta,qstarround,'upper');
avBitDensitytheta     =  gammainc(theta,theta,'lower');
stdBitDensitytheta    =  1/KT*gammainc(theta,theta,'lower').*gammainc(theta,theta,'upper');
set(groot, 'defaultAxesTickLabelInterpreter','latex'); set(groot, 'defaultLegendInterpreter','latex');
figure('Color',[1 1 1],'Position',[81 333 1037 634])
plot(theta,avBitDensitymyqs,'linewidth',2,'color',[0 0 1]);hold on
line([min(theta) max(theta)],[0.6303 0.6302],'LineWidth',2,'LineStyle','--','color',[1 0 0]);text(1,0.64,'$(0.995,0.630)$')
line([min(theta) max(theta)],[0.5 0.5],'LineWidth',2,'LineStyle','--','color',[1 0 0])
line([min(theta) max(theta)],[0.2643 0.2643],'LineWidth',2,'LineStyle','--','color',[1 0 0]);text(1,0.25,'$(1.000,0.264)$')
xlim([min(theta) max(theta)])
xlabel('Exposure $\theta=\alpha c / K$');
ylabel('Average Bit Density $1-E[\gamma_q(c)]$ at $q^*(c)$');
set(gca,'fontsize',20,'fontweight','bold','LineWidth',2)
box off

%% Width of Phase Transition vs KT 
KT       =  1:1000;
delta    =  [1e-3 1e-4 1e-5 1e-6];
theta    =  50;
Psi      =  real(gammainc(theta,Qset,'upper'));
SRIWidth =  zeros(length(KT),1);
for i=1:length(KT)
    for j=1:length(delta)
        Epsilon  =  1-delta(j).^(1./KT(i));
        qrange   =  Qset(Psi>=Epsilon & Psi<=1-Epsilon);
        try
            SRIWidth(i,j) = qrange(end)-qrange(1);
        catch
            SRIWidth(i,j) = 0;
        end
    end
end
set(groot, 'defaultAxesTickLabelInterpreter','latex'); set(groot, 'defaultLegendInterpreter','latex');
figure;plot(KT,SRIWidth,'linewidth',3);
xlabel('Spatio-Temporal Oversampling $KT$');
ylabel('Width of $\mathcal{Q}_\theta$');
legend('$\delta=10^{-3}$','$\delta=10^{-4}$','$\delta=10^{-5}$','$\delta=10^{-6}$')
set(gca,'fontsize',20,'fontweight','bold','LineWidth',2)
title(sprintf('$\\epsilon=1-\\delta^{1/KT}, \\theta=%d$',theta),'interpreter','latex')

%% Boundaries of Phase Transition for different K and T
delta = 2e-4;
for k=1:Kcar
    for t=1:Tcar
        KT                    =  Ksqset(k)*Tset(t);
        Epsilon               =  1-(delta/2)^(1/KT);
        theta                 =  alpha*c0/Ksqset(k);
        Psi                   =  real(gammainc(theta,Qset,'upper'));
        qidxpsi               =  find( Psi >= Epsilon & Psi <= 1-Epsilon );
        p0                    =  Psi.^KT;
        p1                    =  (1-Psi).^KT;
        pSuccess              =  1 - p0 - p1;
        qidxprob              =  find( pSuccess >= 1-delta );
        if ~isempty(qidxprob)
            qlprob(k,t)       =  Qset(qidxprob(1));
            qhprob(k,t)       =  Qset(qidxprob(end));
            qcprob(k,t)       =  0.5*(qlprob(k,t)+qhprob(k,t));
        else
            qlprob(k,t)       =  nan;
            qhprob(k,t)       =  nan;
            qcprob(k,t)       =  nan;
        end
        if ~isempty(qidxpsi)
            qlpsi(k,t)        =  Qset(qidxpsi(1));
            qhpsi(k,t)        =  Qset(qidxpsi(end));
            qcpsi(k,t)        =  0.5*(qlpsi(k,t)+qhpsi(k,t));
        else
            qlpsi(k,t)        =  nan;
            qhpsi(k,t)        =  nan;
            qcpsi(k,t)        =  nan;
        end
    end
end

%% Phase Transition vs K for different T
figure;
plot(2:10,qlprob(:,11),'-b','linewidth',2,'markersize',10);hold on;
plot(2:10,qlpsi(:,11),':b','linewidth',2,'markersize',10);
plot(2:10,qlprob(:,2),'-g','linewidth',2,'markersize',10);
plot(2:10,qlpsi(:,2),':g','linewidth',2,'markersize',10);
plot(2:10,qhprob(:,11),'-b','linewidth',2,'markersize',10);
plot(2:10,qhpsi(:,11),':b','linewidth',2,'markersize',10);
plot(2:10,qhprob(:,2),'-g','linewidth',2,'markersize',10);
plot(2:10,qhpsi(:,2),':g','linewidth',2,'markersize',10);

qstar  =  floor(alpha*c0./Ksqset)+1;
plot(2:10,qstar,'-kp','linewidth',2,'markersize',10);
xlabel('Spatial Oversampling $K$')
xlim([1.8 10.2])
ylabel('Threshold $q$')
set(gca,'XTick',2:10)  
set(gca,'XTicklabels',{'4','9','16','25','36','49','64','81','100'})  
legend('$T=100,\; q\in\{q:P_s>1-\delta\}$','$T=100,\; q\in\Omega_\theta$','$T=10,\; q\in\{q:P_s>1-\delta\}$','$T=10,\; q\in\Omega_\theta$')
set(gca,'fontsize',20,'fontweight','bold','LineWidth',2)
box off

%% Psi function vs theta
epsilon    =  0.01;
theta1     =  0:0.1:20;
x          =  gammainc(theta1,8,'upper');
Thetaq     =  theta1(x<=1-epsilon & x>=epsilon);

figure('Color',[1 1 1],'Position',[95 249 987 676])
set(gca,'fontsize',18,'fontweight','bold','TickLabelInterpreter','latex')
set(groot, 'defaultAxesTickLabelInterpreter','latex'); set(groot, 'defaultLegendInterpreter','latex');
h1 = patch([0 Thetaq(1) Thetaq(1) 0],[epsilon epsilon 1-epsilon 1-epsilon],[1 0 0]);set(h1,'FaceAlpha',0.2,'EdgeColor',[1 1 1]);hold on
h2 = patch([Thetaq(1) Thetaq(end) Thetaq(end) Thetaq(1)],[epsilon epsilon 1-epsilon 1-epsilon],[0 1 0]);set(h2,'FaceAlpha',0.2,'EdgeColor',[1 1 1]);
h3 = patch([Thetaq(end) theta1(end) theta1(end) Thetaq(end)],[epsilon epsilon 1-epsilon 1-epsilon],[1 0 0]);set(h3,'FaceAlpha',0.2,'EdgeColor',[1 1 1]);
h4 = plot(theta1,gammainc(theta1,1,'upper'),'-k','linewidth',3);
h5 = plot(theta1,gammainc(theta1,8,'upper'),'-.b','linewidth',3);
h6 = plot(theta1,gammainc(theta1,16,'upper'),':r','linewidth',3);
xlabel('Intensity $\theta$');ylabel('$\Psi_q(\theta)$');xlim([0 20])
legend([h2 h1 h4 h5 h6],{'$\theta\in\Theta_q, q=8$','$\theta\notin\Theta_q, q=8$','$\Psi_q(\theta), q=1$','$\Psi_q(\theta), q=8$','$\Psi_q(\theta), q=16$'},4)
set(get(gca,'YLabel'),'Rotation',0)
set(gca,'fontsize',14,'fontweight','bold')
set(gcf,'color','w');

%% Psi function vs q
q2         =  1:20;
q2         =  kron(q2,ones(1,100));
x          =  gammainc(8,q2,'upper');
qtheta     =  q2(x<=1-epsilon & x>=epsilon);
ql         =  qtheta(1)*100;
qh         =  qtheta(end)*100;

figure('Color',[1 1 1],'Position',[95 249 987 676])
set(gca,'fontsize',18,'fontweight','bold','TickLabelInterpreter','latex')
set(groot, 'defaultAxesTickLabelInterpreter','latex'); set(groot, 'defaultLegendInterpreter','latex');
msize  =  12;
h1     =  patch([0 ql ql 0],[epsilon epsilon 1-epsilon 1-epsilon],[1 0 0]);set(h1,'FaceAlpha',0.2,'EdgeColor',[1 1 1]);hold on
h2     =  patch([ql qh qh ql],[epsilon epsilon 1-epsilon 1-epsilon],[0 1 0]);set(h2,'FaceAlpha',0.2,'EdgeColor',[1 1 1]);
h3     =  patch([qh q2(end)*100 q2(end)*100 qh],[epsilon epsilon 1-epsilon 1-epsilon],[1 0 0]);set(h3,'FaceAlpha',0.2,'EdgeColor',[1 1 1]);
h4     =  plot(gammainc(1,q2,'upper'),'-k','linewidth',3,'markersize',msize);
h5     =  plot(gammainc(8,q2,'upper'),'-.b','linewidth',3,'markersize',msize);
h6     =  plot(gammainc(16,q2,'upper'),':r','linewidth',3,'markersize',msize);
xlabel('Threshold $q$');ylabel('$\Psi_q(\theta)$')
legend([h2 h1 h4 h5 h6],{'$q\in\Omega_\theta, \theta=8$','$q\notin\Omega_\theta, \theta=8$','$\Psi_q(\theta), \theta=1$','$\Psi_q(\theta), \theta=8$','$\Psi_q(\theta), \theta=16$'},3)
set(get(gca,'YLabel'),'Rotation',0)
set(gca,'fontsize',14,'fontweight','bold')
set(gcf,'color','w');
box off
xlim([1 q2(end)*100])