%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Code to Draw the SNR curves for QIS
%
% O. A. Elgendy and S. H. Chan,
% "Optimal Threshold Design for Quanta Image Sensor," 
% IEEE Transactions on Computational Imaging, 2017.
% http://ieeexplore.ieee.org/document/8169093/
% 
% Omar Elgendy and Stanley Chan
% Copyright 2017
% Purdue University, West Lafayette, In, USA.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

clc,clear,close all
addpath(genpath('./utilities'));

%% Draw Fig. 2a in Supplementary 
c                    =  logspace(-6,0,1e4);   % Normalized Light Coeff.
alpha                =  400;                  % Sensor Gain
K                    =  4;                    % Spatial Oversampling Factor
T                    =  30;                   % Temporal Oversampling Factor
theta                =  alpha * c / K;          % Light Exposure Value

for q=1:16
    SNR_v(:,q)       =  10*log10(K*T)+10*log10(gammainc(theta,q,'lower'))...
                        -10*log10(gammainc(theta,q,'upper'));
end
figure;
for q=1:16
    tone             =  (q-1)*0.05;
    semilogx(c,SNR_v(:,q),'linewidth',2,'Color',tone*[1,1,1]);hold on
end
set(gca,'fontsize',18,'fontweight','bold','TickLabelInterpreter','latex')
set(groot, 'defaultAxesTickLabelInterpreter','latex'); set(groot, 'defaultLegendInterpreter','latex');
xlabel('$c$');ylabel('$\textrm{SNR}_{v,q}(c)$');
ylim([-100 100])

%% Draw Fig. 2b in Supplementary (Fig. 4) in main manuscript

for q=1:16
    SNR(:,q)         =  10/log(10)*(-2*theta+2*q*log(theta)-log(gammainc(theta,q,'upper'))...
                        -log(gammainc(theta,q,'lower'))-2*log(gamma(q)))+10*log10(K*T);
end
figure;
for q=1:16
    tone             =  (q-1)*0.05;
    semilogx(c,SNR(:,q),'linewidth',2,'Color',tone*[1,1,1]);hold on
end
ylim([-100 40])
set(gca,'fontsize',18,'fontweight','bold','TickLabelInterpreter','latex')
set(groot, 'defaultAxesTickLabelInterpreter','latex'); set(groot, 'defaultLegendInterpreter','latex');
xlabel('$c$');ylabel('$\textrm{SNR}_q(c)$');

%% Draw Fig. 8 in main manuscript
theta                =  logspace(-2,7,1e4)';
Q                    =  1:25;
qmax                 =  max(Q);
[J1,J2,J3,J4,LB,UB]  =  deal(zeros(length(theta),length(Q)));
K                    =  5*5;
T                    =  16;
for i=1:length(Q)
    q                =  Q(i);
    J1(:,i)          =  10/log(10)*(-2*theta+2*q*log(theta)-log(real(gammainc(theta,q,'upper')))-log(real(gammainc(theta,q,'lower')))-2*log(gamma(q)))+10*log10(K*T);
    J2(:,i)          =  10/log(10)*(-2*theta*0.2+2*q*log(theta*0.2)-log(real(gammainc(theta*0.2,q,'upper')))-log(real(gammainc(theta*0.2,q,'lower')))-2*log(gamma(q)))+10*log10(K*T);
    J3(:,i)          =  10/log(10)*(-2*theta*0.04+2*q*log(theta*0.04)-log(real(gammainc(theta*0.04,q,'upper')))-log(real(gammainc(theta*0.04,q,'lower')))-2*log(gamma(q)))+10*log10(K*T);
    J4(:,i)          =  10/log(10)*(-2*theta*0.008+2*q*log(theta*0.008)-log(real(gammainc(theta*0.008,q,'upper')))-log(real(gammainc(theta*0.008,q,'lower')))-2*log(gamma(q)))+10*log10(K*T);
    LB(:,i)          =  10/log(10)*(-2*theta+2*q*log(theta)-2*log(gamma(q)));
    UB(:,i)          =  10/log(10)*(-theta/2+log(theta)+(1-q)*(exp(theta).*expint(theta)-1));
end

J1ql                 =  J1(:,1);
J2ql                 =  J2(:,1);
J3ql                 =  J3(:,1);
J4ql                 =  J4(:,1);
x1                   =  10.^(J1ql/10);
x2                   =  10.^(J2ql/10);
x3                   =  10.^(J3ql/10);
x4                   =  10.^(J4ql/10);
Jq1HDR               =  zeros(length(theta),1);
Jq1HDR(x1~=inf)      =  Jq1HDR(x1~=inf)+x1(x1~=inf);
Jq1HDR(x2~=inf)      =  Jq1HDR(x2~=inf)+x2(x2~=inf);
Jq1HDR(x3~=inf)      =  Jq1HDR(x3~=inf)+x3(x3~=inf);
Jq1HDR(x4~=inf)      =  Jq1HDR(x4~=inf)+x4(x4~=inf);
Jq1HDR               =  10*log10(Jq1HDR);

figure('Color',[1 1 1],'Position',[90 250 1000 700])
semilogx(theta,J1ql,'-k','linewidth',2);hold on
semilogx(theta,J2ql,'-.k','linewidth',2);
semilogx(theta,J3ql,':k','linewidth',2);
semilogx(theta,J4ql,'--k','linewidth',2);
semilogx(theta,Jq1HDR,'-k','linewidth',4);

J1qh                 =  J1(:,max(Q));
J2qh                 =  J2(:,max(Q));
J3qh                 =  J3(:,max(Q));
J4qh                 =  J4(:,max(Q));
x1                   =  10.^(J1qh/10);
x2                   =  10.^(J2qh/10);
x3                   =  10.^(J3qh/10);
x4                   =  10.^(J4qh/10);
JqhHDR               =  zeros(length(theta),1);
JqhHDR(x1~=inf)      =  JqhHDR(x1~=inf)+x1(x1~=inf);
JqhHDR(x2~=inf)      =  JqhHDR(x2~=inf)+x2(x2~=inf);
JqhHDR(x3~=inf)      =  JqhHDR(x3~=inf)+x3(x3~=inf);
JqhHDR(x4~=inf)      =  JqhHDR(x4~=inf)+x4(x4~=inf);
JqhHDR               =  10*log10(JqhHDR);

semilogx(theta,J1qh,'-r','linewidth',2);
semilogx(theta,J2qh,'-.r','linewidth',2);
semilogx(theta,J3qh,':r','linewidth',2);
semilogx(theta,J4qh,'--r','linewidth',2);
semilogx(theta,JqhHDR,'-r','linewidth',4);

qstar                =  min(floor(theta*1)+1,qmax);
idx                  =  sub2ind(size(J1), (1:length(theta))', qstar);
J1bestq              =  J1(idx);
  
qstar                =  min(floor(theta*0.2)+1,qmax);
idx                  =  sub2ind(size(J2), (1:length(theta))', qstar);
J2bestq              =  J2(idx);

qstar                =  min(floor(theta*0.04)+1,qmax);
idx                  =  sub2ind(size(J3), (1:length(theta))', qstar);
J3bestq              =  J3(idx);

qstar                =  min(floor(theta*0.008)+1,qmax);
idx                  =  sub2ind(size(J4), (1:length(theta))', qstar);
J4bestq              =  J4(idx);

x1                   =  10.^(J1bestq/10);
x2                   =  10.^(J2bestq/10);
x3                   =  10.^(J3bestq/10);
x4                   =  10.^(J4bestq/10);
JbestqHDR            =  zeros(length(theta),1);
JbestqHDR(x1~=inf)   =  JbestqHDR(x1~=inf)+x1(x1~=inf);
JbestqHDR(x2~=inf)   =  JbestqHDR(x2~=inf)+x2(x2~=inf);
JbestqHDR(x3~=inf)   =  JbestqHDR(x3~=inf)+x3(x3~=inf);
JbestqHDR(x4~=inf)   =  JbestqHDR(x4~=inf)+x4(x4~=inf);
JbestqHDR            =  10*log10(JbestqHDR);

semilogx(theta,J1bestq,'-b','linewidth',2);
semilogx(theta,J2bestq,'-.b','linewidth',2);
semilogx(theta,J3bestq,':b','linewidth',2);
semilogx(theta,J4bestq,'-b','linewidth',2);
semilogx(theta,JbestqHDR,'-b','linewidth',4);
ylim([-30 60]);xlim([1e-2 1e4])
xlabel('$\theta=\alpha c / K^2$');ylabel('$SNR_q(c)$');set(gca,'fontsize',20,'fontweight','bold');%set(get(gca,'YLabel'),'Rotation',0);
legend_str=[{'$q=1, \tau=1$'},{'$q=1, \tau=0.2$'},{'$q=1, \tau=0.04$'},{'$q=1, \tau=0.008$'},{'$q=1$, overall SNR'},...
            {'$q=25, \tau=1$'},{'$q=25, \tau=0.2$'},{'$q=25, \tau=0.04$'},{'$q=25, \tau=0.008$'},{'$q=25$, overall SNR'},...
            {'$q^{\ast}(c), \tau=1$'},{'$q^{\ast}(c), \tau=0.2$'},{'$q^{\ast}(c), \tau=0.04$'},{'$q^{\ast}(c), \tau=0.008$'},{'$q^{\ast}(c)$, overall SNR'}];
{[sprintf('$q=%d,',Q(end)) ' \tau=1']},{[sprintf('q=%d,',Q(end)) ' \tau=0.2']},{[sprintf('q=%d,',Q(end)) ' \tau=0.04']},{[sprintf('q=%d,',Q(end)) ' \tau=0.08']},{[sprintf('q=%d,',Q(end)) ' overall SNR']},...
set(gca,'fontsize',18,'fontweight','bold','TickLabelInterpreter','latex')
set(groot, 'defaultAxesTickLabelInterpreter','latex'); set(groot, 'defaultLegendInterpreter','latex');
h=columnlegend(3, legend_str, 'Location', 'best');
set(gca,'fontsize',18,'fontweight','bold','TickLabelInterpreter','latex')
set(groot, 'defaultAxesTickLabelInterpreter','latex'); set(groot, 'defaultLegendInterpreter','latex');
set(gca, 'box', 'off')